<?php
/**
 * Copyright (C) 2017-2019 thirty bees
 * Copyright (C) 2007-2016 PrestaShop
 *
 * thirty bees is an extension to the PrestaShop software by PrestaShop SA.
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.md.
 * It is also available through the world-wide-web at this URL:
 * https://opensource.org/licenses/osl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@thirtybees.com so we can send you a copy immediately.
 *
 * @author    thirty bees <contact@thirtybees.com>
 * @author    PrestaShop SA <contact@prestashop.com>
 * @copyright 2017-2019 thirty bees
 * @copyright 2007-2016 PrestaShop SA
 * @license   https://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 * PrestaShop is an internationally registered trademark of PrestaShop SA.
 */

namespace TbUpdaterModule;

/**
 * Class GroupCore
 *
 * @since 1.0.0
 */
class Group extends ObjectModel
{
    // @codingStandardsIgnoreStart
    protected static $cache_reduction = [];
    protected static $group_price_display_method = [];
    /** @var string Lastname */
    public $name;
    /** @var string Reduction */
    public $reduction;
    /** @var int Price display method (tax inc/tax exc) */
    public $price_display_method;
    /** @var bool Show prices */
    public $show_prices = 1;
    /** @var string Object creation date */
    public $date_add;
    /** @var string Object last modification date */
    public $date_upd;
    // @codingStandardsIgnoreEnd

    /**
     * @see ObjectModel::$definition
     */
    public static $definition = [
        'table'     => 'group',
        'primary'   => 'id_group',
        'multilang' => true,
        'fields'    => [
            'reduction'            => ['type' => self::TYPE_FLOAT, 'validate' => 'isFloat'],
            'price_display_method' => ['type' => self::TYPE_INT, 'validate' => 'isPriceDisplayMethod', 'required' => true],
            'show_prices'          => ['type' => self::TYPE_BOOL, 'validate' => 'isBool'],
            'date_add'             => ['type' => self::TYPE_DATE, 'validate' => 'isDate'],
            'date_upd'             => ['type' => self::TYPE_DATE, 'validate' => 'isDate'],

            /* Lang fields */
            'name'                 => ['type' => self::TYPE_STRING, 'lang' => true, 'validate' => 'isGenericName', 'required' => true, 'size' => 32],
        ],
    ];

    protected $webserviceParameters = [];

    /**
     * GroupCore constructor.
     *
     * @param int|null $id
     * @param int|null $idLang
     * @param int|null $idShop
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public function __construct($id = null, $idLang = null, $idShop = null)
    {
        parent::__construct($id, $idLang, $idShop);
        if ($this->id && !isset(Group::$group_price_display_method[$this->id])) {
            static::$group_price_display_method[$this->id] = $this->price_display_method;
        }
    }

    /**
     * @param int      $idLang
     * @param int|bool $idShop
     *
     * @return array|false|mysqli_result|null|PDOStatement|resource
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function getGroups($idLang, $idShop = false)
    {
        $shopCriteria = '';
        if ($idShop) {
            $shopCriteria = Shop::addSqlAssociation('group', 'g');
        }

        return Db::getInstance(_PS_USE_SQL_SLAVE_)->executeS(
            '
		SELECT DISTINCT g.`id_group`, g.`reduction`, g.`price_display_method`, gl.`name`
		FROM `'._DB_PREFIX_.'group` g
		LEFT JOIN `'._DB_PREFIX_.'group_lang` AS gl ON (g.`id_group` = gl.`id_group` AND gl.`id_lang` = '.(int) $idLang.')
		'.$shopCriteria.'
		ORDER BY g.`id_group` ASC'
        );
    }

    /**
     * @param null $idCustomer
     *
     * @return mixed
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function getReduction($idCustomer = null)
    {
        if (!isset(static::$cache_reduction['customer'][(int) $idCustomer])) {
            $idGroup = $idCustomer ? Customer::getDefaultGroupId((int) $idCustomer) : (int) Group::getCurrent()->id;
            static::$cache_reduction['customer'][(int) $idCustomer] = Group::getReductionByIdGroup($idGroup);
        }

        return static::$cache_reduction['customer'][(int) $idCustomer];
    }

    /**
     * Return current group object
     * Use context
     *
     * @return Group Group object
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function getCurrent()
    {
        static $groups = [];
        static $psUnidentifiedGroup = null;
        static $psCustomerGroup = null;

        if ($psUnidentifiedGroup === null) {
            $psUnidentifiedGroup = Configuration::get('PS_UNIDENTIFIED_GROUP');
        }

        if ($psCustomerGroup === null) {
            $psCustomerGroup = Configuration::get('PS_CUSTOMER_GROUP');
        }

        $customer = Context::getContext()->customer;
        if (Validate::isLoadedObject($customer)) {
            $idGroup = (int) $customer->id_default_group;
        } else {
            $idGroup = (int) $psUnidentifiedGroup;
        }

        if (!isset($groups[$idGroup])) {
            $groups[$idGroup] = new Group($idGroup);
        }

        if (!$groups[$idGroup]->isAssociatedToShop(Context::getContext()->shop->id)) {
            $idGroup = (int) $psCustomerGroup;
            if (!isset($groups[$idGroup])) {
                $groups[$idGroup] = new Group($idGroup);
            }
        }

        return $groups[$idGroup];
    }

    /**
     * @param int $idGroup
     *
     * @return mixed
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function getReductionByIdGroup($idGroup)
    {
        if (!isset(static::$cache_reduction['group'][$idGroup])) {
            static::$cache_reduction['group'][$idGroup] = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue(
                '
			SELECT `reduction`
			FROM `'._DB_PREFIX_.'group`
			WHERE `id_group` = '.(int) $idGroup
            );
        }

        return static::$cache_reduction['group'][$idGroup];
    }

    /**
     * @return mixed
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function getDefaultPriceDisplayMethod()
    {
        return Group::getPriceDisplayMethod((int) Configuration::get('PS_CUSTOMER_GROUP'));
    }

    /**
     * @param $idGroup
     *
     * @return mixed
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function getPriceDisplayMethod($idGroup)
    {
        if (!isset(Group::$group_price_display_method[$idGroup])) {
            static::$group_price_display_method[$idGroup] = Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue(
                '
			SELECT `price_display_method`
			FROM `'._DB_PREFIX_.'group`
			WHERE `id_group` = '.(int) $idGroup
            );
        }

        return static::$group_price_display_method[$idGroup];
    }

    /**
     * This method is allow to know if a feature is used or active
     *
     * @return bool
     *
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function isFeatureActive()
    {
        static $psGroupFeatureActive = null;
        if ($psGroupFeatureActive === null) {
            $psGroupFeatureActive = Configuration::get('PS_GROUP_FEATURE_ACTIVE');
        }

        return $psGroupFeatureActive;
    }

    /**
     * This method is allow to know if there are other groups than the default ones
     *
     * @param string $table
     * @param bool   $hasActiveColumn
     *
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function isCurrentlyUsed($table = null, $hasActiveColumn = false)
    {
        return (bool) (Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue('SELECT COUNT(*) FROM `'._DB_PREFIX_.'group`') > 3);
    }

    /**
     * Truncate all restrictions by module
     *
     * @param int $idModule
     *
     * @return bool
     *
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function truncateRestrictionsByModule($idModule)
    {
        return Db::getInstance()->execute(
            '
		DELETE FROM `'._DB_PREFIX_.'module_group`
		WHERE `id_module` = '.(int) $idModule
        );
    }

    /**
     * Adding restrictions modules to the group with id $id_group
     *
     * @param int   $idGroup
     * @param array $modules
     * @param array $shops
     *
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function addModulesRestrictions($idGroup, $modules, $shops = [1])
    {
        if (!is_array($modules) || !count($modules) || !is_array($shops) || !count($shops)) {
            return false;
        }

        // Delete all record for this group
        Db::getInstance()->execute('DELETE FROM `'._DB_PREFIX_.'module_group` WHERE `id_group` = '.(int) $idGroup);

        $sql = 'INSERT INTO `'._DB_PREFIX_.'module_group` (`id_module`, `id_shop`, `id_group`) VALUES ';
        foreach ($modules as $module) {
            foreach ($shops as $shop) {
                $sql .= '("'.(int) $module.'", "'.(int) $shop.'", "'.(int) $idGroup.'"),';
            }
        }
        $sql = rtrim($sql, ',');

        return (bool) Db::getInstance()->execute($sql);
    }

    /**
     * Add restrictions for a new module.
     * We authorize every groups to the new module
     *
     * @param int   $idModule
     * @param array $shops
     *
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function addRestrictionsForModule($idModule, $shops = [1])
    {
        if (!is_array($shops) || !count($shops)) {
            return false;
        }

        $res = true;
        foreach ($shops as $shop) {
            $res &= Db::getInstance()->execute(
                '
			INSERT INTO `'._DB_PREFIX_.'module_group` (`id_module`, `id_shop`, `id_group`)
			(SELECT '.(int) $idModule.', '.(int) $shop.', id_group FROM `'._DB_PREFIX_.'group`)'
            );
        }

        return $res;
    }

    /**
     * Light back office search for Group
     *
     * @param string $query Searched string
     *
     * @return array Corresponding groups
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function searchByName($query)
    {
        return Db::getInstance()->getRow(
            '
			SELECT g.*, gl.*
			FROM `'._DB_PREFIX_.'group` g
			LEFT JOIN `'._DB_PREFIX_.'group_lang` gl
				ON (g.`id_group` = gl.`id_group`)
			WHERE `name` = \''.pSQL($query).'\'
		'
        );
    }

    /**
     * @param bool $count
     * @param int  $start
     * @param int  $limit
     * @param bool $shopFiltering
     *
     * @return array|false|mysqli_result|null|PDOStatement|resource|string
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public function getCustomers($count = false, $start = 0, $limit = 0, $shopFiltering = false)
    {
        if ($count) {
            return Db::getInstance()->getValue(
                '
			SELECT COUNT(*)
			FROM `'._DB_PREFIX_.'customer_group` cg
			LEFT JOIN `'._DB_PREFIX_.'customer` c ON (cg.`id_customer` = c.`id_customer`)
			WHERE cg.`id_group` = '.(int) $this->id.'
			'.($shopFiltering ? Shop::addSqlRestriction(Shop::SHARE_CUSTOMER) : '').'
			AND c.`deleted` != 1'
            );
        }

        return Db::getInstance()->executeS(
            '
		SELECT cg.`id_customer`, c.*
		FROM `'._DB_PREFIX_.'customer_group` cg
		LEFT JOIN `'._DB_PREFIX_.'customer` c ON (cg.`id_customer` = c.`id_customer`)
		WHERE cg.`id_group` = '.(int) $this->id.'
		AND c.`deleted` != 1
		'.($shopFiltering ? Shop::addSqlRestriction(Shop::SHARE_CUSTOMER) : '').'
		ORDER BY cg.`id_customer` ASC
		'.($limit > 0 ? 'LIMIT '.(int) $start.', '.(int) $limit : '')
        );
    }

    /**
     * @param bool $autodate
     * @param bool $nullValues
     *
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public function add($autodate = true, $nullValues = false)
    {
        Configuration::updateGlobalValue('PS_GROUP_FEATURE_ACTIVE', '1');
        if (parent::add($autodate, $nullValues)) {
            Category::setNewGroupForHome((int) $this->id);
            Carrier::assignGroupToAllCarriers((int) $this->id);

            return true;
        }

        return false;
    }

    /**
     * @param bool $autodate
     * @param bool $nullValues
     *
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public function update($autodate = true, $nullValues = false)
    {
        if (!Configuration::getGlobalValue('PS_GROUP_FEATURE_ACTIVE') && $this->reduction > 0) {
            Configuration::updateGlobalValue('PS_GROUP_FEATURE_ACTIVE', 1);
        }

        return parent::update($autodate, $nullValues);
    }

    /**
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public function delete()
    {
        if ($this->id == (int) Configuration::get('PS_CUSTOMER_GROUP')) {
            return false;
        }
        if (parent::delete()) {
            Db::getInstance()->execute('DELETE FROM `'._DB_PREFIX_.'cart_rule_group` WHERE `id_group` = '.(int) $this->id);
            Db::getInstance()->execute('DELETE FROM `'._DB_PREFIX_.'customer_group` WHERE `id_group` = '.(int) $this->id);
            Db::getInstance()->execute('DELETE FROM `'._DB_PREFIX_.'category_group` WHERE `id_group` = '.(int) $this->id);
            Db::getInstance()->execute('DELETE FROM `'._DB_PREFIX_.'group_reduction` WHERE `id_group` = '.(int) $this->id);
            Db::getInstance()->execute('DELETE FROM `'._DB_PREFIX_.'product_group_reduction_cache` WHERE `id_group` = '.(int) $this->id);
            $this->truncateModulesRestrictions($this->id);

            // Add default group (id 3) to customers without groups
            Db::getInstance()->execute(
                'INSERT INTO `'._DB_PREFIX_.'customer_group` (
				SELECT c.id_customer, '.(int) Configuration::get('PS_CUSTOMER_GROUP').' FROM `'._DB_PREFIX_.'customer` c
				LEFT JOIN `'._DB_PREFIX_.'customer_group` cg
				ON cg.id_customer = c.id_customer
				WHERE cg.id_customer IS NULL)'
            );

            // Set to the customer the default group
            // Select the minimal id from customer_group
            Db::getInstance()->execute(
                'UPDATE `'._DB_PREFIX_.'customer` cg
				SET id_default_group =
					IFNULL((
						SELECT min(id_group) FROM `'._DB_PREFIX_.'customer_group`
						WHERE id_customer = cg.id_customer),
						'.(int) Configuration::get('PS_CUSTOMER_GROUP').')
				WHERE `id_default_group` = '.(int) $this->id
            );

            return Db::getInstance()->delete('module_group', '`id_group` = '.(int) $this->id);
        }

        return false;
    }

    /**
     * Truncate all modules restrictions for the group
     *
     * @param int $idGroup
     *
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public static function truncateModulesRestrictions($idGroup)
    {
        return Db::getInstance()->delete(
            'module_group',
            '`id_group` = '.(int) $idGroup
        );
    }
}
